package ch.epfl.lamp.minischeme;

import java.io.PrintStream;
import scala.collection.immutable.{Set,ListSet,Map,ListMap};

/**
 * Code generator for minischeme.
 *
 * @author Michel Schinz
 */
object Generator extends Function2[List[Tree],PrintStream,Unit] {
  def apply(trees: List[Tree], stream: PrintStream): Unit = {
    val code: Code = new Code();

    // Attribute locations to global symbols
    var globalOffset: Int = 0;
    trees foreach {
      case SList(SIdent("define") :: name :: value :: Nil) =>
        value match {
          case SList(SIdent("lambda") :: _) =>
            name.symbol.location = FunLocation(code.newLabel(name.symbol.name))
          case _ =>
            name.symbol.location = GlobalLocation(globalOffset);
            globalOffset = globalOffset + code.WORD_SIZE
        }
      case _ =>
    }

    // Allocate space for global variables
    code.emit(LINT, code.GP, globalOffset);
    code.emit(ALOC, code.GP, code.GP);

    // Allocate global "stack frame"
    val loadGlobalFrameSizeInstr =
      new InstructionRC(LINT, code.FP, IntegerConstant(0));
    code.emit(loadGlobalFrameSizeInstr);
    code.emit(ALOC, code.FP, code.FP);

    val Pair(funDefs, exprs) = trees partition {
      case SList(SIdent("define"):: _ :: SList(SIdent("lambda") :: _) :: Nil) =>
        true
      case _ =>
        false
    }
    // Generate expressions
    exprs foreach { e => gen(code, e) };
    loadGlobalFrameSizeInstr.constant = IntegerConstant(code.getMaxFrameSize);
    code.resetMaxFrameSize;
    code.emit(HALT);

    // generate function bodies
    funDefs foreach { f => gen(code, f) };

    code dumpTo stream
  }

  private def gen(code: Code, tree: Tree): Unit = tree match {
    case SList(SIdent("define") :: name :: SList(SIdent("lambda") :: SList(args) :: body) :: Nil) =>
      assert(code.getFrameSize == 0);

      code.addComment("begin " + name.symbol.name);
      name.symbol.location match {
        case FunLocation(label) => code.emit(label)
      }
      for (val argReg <- Iterator.range(1, args.length + 1))
        code.allocateRegister(code.registers(argReg));
      // allocate and link frame
      code.resetMaxFrameSize;
      var loadFrameSizeInstr: InstructionRC = null;
      code withFreshRegister { tmpReg =>
        loadFrameSizeInstr = new InstructionRC(LINT, tmpReg, IntegerConstant(0));
        code.emit(loadFrameSizeInstr);
        code.emit(ALOC, tmpReg, tmpReg);
        code.emit(STOR, code.FP, tmpReg, 0);
        code.incFrameSize(code.WORD_SIZE);
        code.emit(CMOV, code.FP, tmpReg, code.R0);
      };
      // save all parameters (including LK)
      code.emitPush(code.LK);
      for (val Pair(reg, arg) <- List.range(1, args.length + 1) zip args) {
        arg.symbol.location = FrameLocation(code.getFrameSize);
        code.emitPush(code.registers(reg));
        code.freeRegister(code.registers(reg));
      }
      val retValReg = code.allocateRegister(code.registers(1));
      for (val statement <- body)
        genLoad(code, statement, retValReg);
      code.freeRegister(retValReg);
      code.decFrameSize(args.length * code.WORD_SIZE);
      code.emitPop(code.LK);
      code.emitPop(code.FP);
      assert(code.getFrameSize == 0, "non-zero frame size");
      code.emit(CMOV, code.PC, code.LK, code.R0); // return
      loadFrameSizeInstr.constant = IntegerConstant(code.getMaxFrameSize);
      code.addComment("end " + name.symbol.name)

    case SList(SIdent("define") :: name :: value :: Nil) =>
      code.addComment("begin " + name.symbol.name);
      code withFreshRegister { tmpReg =>
        genLoad(code, value, tmpReg);
        name.symbol.location match {
          case GlobalLocation(o) => code.emit(STOR, tmpReg, code.GP, o);
        }
      };
      code.addComment("end " + name.symbol.name)

    case _ =>
      code withFreshRegister { tmpReg => genLoad(code, tree, tmpReg) }
  }

  private val binPrims = (new ListMap[String, OpcodeRRR]()
    + "$+" -> ADD + "$-" -> SUB
    + "$*" -> MUL + "$/" -> DIV + "$%" -> MOD
    + "$=" -> ISEQ + "$<" -> ISLT + "$<=" -> ISLE);

  private def genLoad(code: Code, tree: Tree, targetReg: Register): Unit =
  tree match {
    case SList(SIdent("if") :: SList(SIdent("$not") :: cond :: Nil) :: thenp :: elsep :: Nil) =>
      genLoad(code,
              SList(SIdent("if") :: cond :: elsep :: thenp :: Nil),
              targetReg)
    case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
      genLoad(code, cond, targetReg);
      val elseLabel = code.newLabel("else");
      code.emitJump(elseLabel, targetReg);
      genLoad(code, thenp, targetReg);
      val afterLabel = code.newLabel;
      code.emitJump(afterLabel, code.R0);
      code.emit(elseLabel);
      genLoad(code, elsep, targetReg);
      code.emit(afterLabel)

    case SList(SIdent("let") :: SList(bindings) :: body) =>
      for (val SList(ident :: value :: Nil) <- bindings) {
        genLoad(code, value, targetReg);
        ident.symbol.location = FrameLocation(code.getFrameSize);
        code.emitPush(targetReg);
      };
      for (val statement <- body)
        genLoad(code, statement, targetReg);
      code.decFrameSize(bindings.length * code.WORD_SIZE)

    case SNum(value) =>
      code.emit(LINT, targetReg, value)

    case SList(SIdent(ident) :: x :: y :: Nil) if (binPrims contains ident) =>
      genLoad(code, x, targetReg);
      code withFreshRegister { yReg =>
        genLoad(code, y, yReg);
        code.emit(binPrims(ident), targetReg, targetReg, yReg);
      }

    case SList(SIdent("$alloc") :: SNum(size) :: Nil) =>
      genLoad(code, SNum(size * code.WORD_SIZE), targetReg);
      code.emit(ALOC, targetReg, targetReg)
    case SList(SIdent("$get") :: block :: SNum(offset) :: Nil) =>
      genLoad(code, block, targetReg);
      code.emit(LOAD, targetReg, targetReg, offset * code.WORD_SIZE)
    case SList(SIdent("$set") :: block :: SNum(offset) :: value :: Nil) =>
      genLoad(code, block, targetReg);
      code withFreshRegister { tmpReg =>
        genLoad(code, value, tmpReg);
        code.emit(STOR, tmpReg, targetReg, offset * code.WORD_SIZE)
      }
    case SList(SIdent("$print-int") :: x :: Nil) =>
      genLoad(code, x, targetReg);
      code.emit(PINT, targetReg)
    case SList(SIdent("$read-int") :: Nil) =>
      code.emit(RINT, targetReg)
    case SList(SIdent("$print-char") :: x :: Nil) =>
      genLoad(code, x, targetReg);
      code.emit(PCHR, targetReg)
    case SList(SIdent("$read-char") :: Nil) =>
      code.emit(RCHR, targetReg)

    case SList(function :: args) =>
      // save registers
      val registersToSave = code.allocatedRegisters diff List(targetReg);
      for (val r <- registersToSave) {
        code.emitPush(r);
        code.freeRegister(r);
      }
      code.freeRegister(targetReg);
      val funReg = code.allocateRegister(code.registers(27));
      genLoad(code, function, funReg);
      for (val i <- Iterator.range(0,args.length)) {
        val reg = code.registers(i + 1);
        code.allocateRegister(reg);
        genLoad(code, args(i), reg);
      }
      val retLabel = code.newLabel("ret");
      code.emit(LINT, code.LK, retLabel);
      code.emit(CMOV, code.PC, funReg, code.R0);
      code.emit(retLabel);
      code.freeRegister(funReg);
      for (val i <- Iterator.range(0,args.length))
        code.freeRegister(code.registers(i + 1));
      val retValReg = code.registers(1);
      code.allocateRegister(targetReg);
      if (targetReg != retValReg)
        code.emit(CMOV, targetReg, retValReg, code.R0);
      // restore registers
      for (val r <- registersToSave.reverse) {
        code.allocateRegister(r);
        code.emitPop(r)
      }

    case SIdent(name) =>
      tree.symbol.location match {
        case FrameLocation(offset) =>
          code.emit(LOAD, targetReg, code.FP, offset);
        case GlobalLocation(offset) =>
          code.emit(LOAD, targetReg, code.GP, offset);
        case FunLocation(label) =>
          code.emit(LINT, targetReg, label);
      }

    case _ =>
      Report.error(tree.line, tree.column, "invalid tree: " + tree);
  }
}
