/*
 * MiniVM project (c) LAMP 2006
 * 
 * Aligned loader of the input file.
 * 
 * $Id: alignloader.c,v 1.2 2006/03/23 16:36:47 dragos Exp $
 */

/*
 * This loader uses the following layout of code_mem:
 * 
 * 00 - the opcode (4 bytes)
 * 04 - the register arguments, 5 bits per register, starting with the
 *      most signifficat bits. (4 bytes)
 * 08 - optional int, for LINT and LOAD/STOAR operations (4 bytes)
 */
 
#include <ctype.h>
 
#include "loader.h"
#include "parser.h"
#include "tokens.h"

#include "../main.h"
#include "../emulator/vm.h"

#define PUT_OPCODE(pc, code) (*((int*) (code_mem + pc))) = (code)

#define MASK_R1 0xF8000000
#define MASK_R2 0x07C00000
#define MASK_R3 0x003E0000

#define GET_R1(w) (((w) & MASK_R1) >> 27)
#define GET_R2(w) (((w) & MASK_R2) >> 22)
#define GET_R3(w) (((w) & MASK_R3) >> 17)

#define SET_R1(pc, val)  ((*(int*)(code_mem + pc)) |= (val << 27))
#define SET_R2(pc, val)  ((*(int*)(code_mem + pc)) |= (val << 22))
#define SET_R3(pc, val)  ((*(int*)(code_mem + pc)) |= (val << 17))

void instruction();

int decode(int pc, int* arg1, int* arg2, int* arg3);
void decode_args(int pc, int opcode, int* arg1, int* arg2, int* arg3);
void next_pc(int opcode);
int instruction_size(int opcode);

/* Parse an instruction and write it in code_mem[pc]. Increment pc. */
void
instruction()
{
  int op = 0;
  
  if (pc > code_size - 3*WORD_SIZE)
  {
    printf("Not enough code memory to load program. Use '--code' to specify more.\n");
    exit_vm();
  }
  
  switch (op = opcode())
    {
    case LINT:
      PUT_OPCODE(pc, OP_LINT);
      pc += WORD_SIZE;
      SET_R1(pc, reg());
      pc += WORD_SIZE;
      skip_whitespace();
      if (isdigit(line[c]))
        {
          int n = integer();
          PUT_WORD(pc, n);
          pc += WORD_SIZE;
        } else
          {
            handle_label();
            pc += WORD_SIZE;
          }
      break;

    case ADD:       case SUB:       case MUL:
    case DIV:       case MOD:       case ISLT:
    case ISLE:      case ISEQ:      case CMOV:
      PUT_OPCODE(pc, TOK2OP(op));
      pc += WORD_SIZE;
      SET_R1(pc, reg());
      SET_R2(pc, reg());
      SET_R3(pc, reg());
      pc += WORD_SIZE;
      break;
                        
    case ALOC:
      PUT_OPCODE(pc, TOK2OP(op));
      pc += WORD_SIZE;      
      SET_R1(pc, reg());
      SET_R2(pc, reg());
      pc += WORD_SIZE;
      break;
                        
    case LOAD:      case STOR:
      PUT_OPCODE(pc, TOK2OP(op));
      pc += WORD_SIZE;
      SET_R1(pc, reg());
      SET_R2(pc, reg());
      pc += WORD_SIZE;
      skip_whitespace();
      if (isdigit(line[c]))
        {
          PUT_WORD(pc, integer());
          pc += WORD_SIZE;
        } else
          {
            handle_label();
            pc += WORD_SIZE;
          }
      break;
                
    case RINT:      case PINT:
    case RCHR:      case PCHR:
      PUT_OPCODE(pc, TOK2OP(op));
      pc += WORD_SIZE;
      SET_R1(pc, reg());
      pc += WORD_SIZE;
      break;
    
    case HALT:
      PUT_OPCODE(pc, TOK2OP(op));
      pc += WORD_SIZE;
      break;
                        
    default:
      error("Unknown opcode.");       
    }
}

/* Return the opcode of the current instruction, given as hard-memory pointer
 * and decode arguments (registers or integers).
 * It does not update the value of R[31] to point to the
 * next instruction. */
int
decode(int pc, int* arg1, int* arg2, int* arg3)
{
  int opcode = *(int*)pc;
  int regs   = *(int*)(pc + 4);
  *arg1 = GET_R1(regs);
  *arg2 = GET_R2(regs);
  *arg3 = GET_R3(regs);
        
  switch (opcode)
    {
      case OP_LINT:   case OP_LOAD:   case OP_STOR:
        *arg3 = GET_WORD(pc + 8);
        break;
    }
  return opcode;  
}

/* Decode the arguments of the given opcode. PC should point to the
 * beginning of the instruction (including the opcode). */
void
decode_args(int pc, int opcode, int* arg1, int* arg2, int* arg3)
{
  int r = *(int*)(pc + 4);
  *arg1 = GET_R1(r);
  *arg2 = GET_R2(r);
  *arg3 = GET_R3(r);
  
  switch (opcode)
    {
      case OP_LINT:   case OP_LOAD:   case OP_STOR:
        *arg3 = GET_WORD(pc + 8);
        break;
      
      case OP_HALT:
        break;
        
      default:
        break;
    }
}


/* Increment the PC to point to the next instruction. */
void
next_pc(int opcode)
{
  R[31] += instruction_size(opcode);
}

/* Return the size in bytes of the given opcode. */
int
instruction_size(int opcode)
{
  /* the length of one instruction is usually 8 bytes */
  int len = 2 * WORD_SIZE;
  switch (opcode)
    {
      case OP_LINT:   case OP_LOAD:   case OP_STOR:
        len += WORD_SIZE;
        break;
      case OP_HALT:
        len = WORD_SIZE;
        break;
    }
  return len;
}
